#
# Copyright 2020 Vianai Systems, Inc. Vianai Confidential & Proprietary Information
#


import sys, os
#sys.path.insert(0, '%s/..'%os.environ['PROJECT_PATH'])

from . import model as vian_model
import lightgbm
import pandas as pd
import numpy as np
import pytest
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler, LabelEncoder, OneHotEncoder
from sklearn import preprocessing
from sklearn.compose import ColumnTransformer
from sklearn.metrics import mean_absolute_error, r2_score, mean_squared_error

from pmdarima.arima import AutoARIMA

from sklearn.metrics import precision_score, recall_score, f1_score, roc_auc_score, accuracy_score, classification_report
from sklearn.preprocessing import LabelBinarizer
import tensorflow as tf
import tensorflow.keras.backend as K
import tensorflow.keras.losses

from sklearn.metrics import mean_absolute_error, r2_score, mean_squared_error, mean_absolute_error
# import keras
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense,Dropout,Embedding,BatchNormalization
from tensorflow.keras.callbacks import EarlyStopping

from sklearn.model_selection import train_test_split
import lightgbm
import shap
from sklearn.preprocessing import MinMaxScaler

class experiment(object):
    def __init__(self, **parameter_dict):
        pass
    def train(self):
        raise NotImplementedError
    def save(self):
        raise NotImplementedError
    def test_save(self):
        raise NotImplementedError
    def configuration(self, **parameter_dict):
        raise NotImplementedError
    def run_experiment(self, **parameter_dict):
        pass

    def configuration_check(self, keys, **parameter_dict):
        for k in keys:
            assert k in parameter_dict, ('parameter %s missing from input'%k)

    def persistent(self, dataframe, filename):
        # dataframe check
        # TODO: type check
        dataframe.to_csv('%s/%s/%s'%(self.model_repo, self.model_desc, filename), index=False)

    def calc_result(self, dataframe, actual, pred):
        """
        Calculate final prediction result
        """
        df_result = pd.DataFrame({'actual':actual, 
                                  'pred':pred
                                 })
                                
        return pd.concat([dataframe[self.additional_persistent_fields].reset_index(drop=True), df_result.reset_index(drop=True)], axis=1)

    def calc_metrics(self):
        raise NotImplementedError
        
    def multiclass_roc_auc_score(self, truth, pred):
        """
        Calculate ROC AUC for multiclass problem
        """
        lb = LabelBinarizer()
        lb.fit(truth)
        truth = lb.transform(truth)
        pred = lb.transform(pred)
        return roc_auc_score(truth, pred)
    

class ImpExperiment(experiment):
    """Regular Regression Model
    """
    def __init__(self, **parameter_dict): 
        self.configuration(**parameter_dict)

    # TODO: abstraction and inheritance
    def configuration(self, **parameter_dict):
        # 
        try:
            self.configuration_check(['estimator', 'cat_cols', 'cont_cols', 'model_repo', 'desc', 'train', 'test', 'target', 'additional_persistent_fields'], **parameter_dict)
        except AssertionError:
            tb = sys.exc_info()[2]
            raise AssertionError('configuration error').with_traceback(tb)

        estimator = parameter_dict['estimator']
        # print(parameter_dict['estimator_param'])
        estimator.set_params(**parameter_dict['estimator_param'])

        self.cat_cols = parameter_dict['cat_cols']
        self.cont_cols = parameter_dict['cont_cols']
        
        self.model_repo = parameter_dict['model_repo']
        self.model_desc = parameter_dict['desc']

        self.model = vian_model.regularModel(parameter_dict['model_repo'], parameter_dict['desc'])
        self.model.set_estimator(estimator)

        self.x_cols = parameter_dict['cat_cols'] + parameter_dict['cont_cols']
        self.y_col = parameter_dict['target']
        self.train_data = parameter_dict['train']
        self.test_data = parameter_dict['test']
        self.valid_data = parameter_dict['validation']
        self.train_x, self.train_y = parameter_dict['train'][self.x_cols], parameter_dict['train'][self.y_col] 
        self.test_x, self.test_y = parameter_dict['test'][self.x_cols], parameter_dict['test'][self.y_col] 
        self.encode_method = parameter_dict['encode_method']

        self.additional_persistent_fields = parameter_dict['additional_persistent_fields']

    def train(self):
        self.model.train(self.train_x, self.train_y)
    def save(self):
        self.model.save()  
    def test_and_save(self):
        self.y_true, self.y_pred = self.test_y, self.model.predict(self.test_x)

        # result persistent
        self.persistent(self.calc_result(self.test_data, self.y_true.values.flatten(), self.y_pred), 'test/test.csv')
        # metrics persistent
        self.persistent(self.calc_metrics(), 'test/metrics.csv')
    def run_experiment(self, **parameter_dict):
        """
        train the model, save model and save test result and metrics
        """
        # training 
        self.train()

        # save model
        self.save()

        self.test_and_save()
        
    def train_shap(self):
        """Train pure lightgbm for shap value generation"""
        if self.encode_method == 'onehot':
            self.cat_cols_trans = self.model.estimator.get_params()['preprocessor']\
                           .transformers_[1][1]['oneHot'].get_feature_names(self.cat_cols)
        elif self.encode_method == 'label':
            self.cat_cols_trans = self.cat_cols

        self.x_cols_new = self.cont_cols + list(self.cat_cols_trans)
        
        
        self.treeExplainer = shap.TreeExplainer(self.model.estimator.get_params()['ml_model'])
        

        self.x_train_trans = self.model.estimator.get_params()['preprocessor'].transform(self.train_x)
        self.x_test_trans = self.model.estimator.get_params()['preprocessor'].transform(self.test_x)
        


        self.shap_values = self.treeExplainer.shap_values(self.x_train_trans)
        self.test_shap_values = self.treeExplainer.shap_values(self.x_test_trans)

        if type(self.shap_values) == list:
            self.shap_values = self.shap_values[1]
            self.test_shap_values = self.test_shap_values[1]

    def generate_shap_train(self):
        """generate shap values for training sets"""
        if self.encode_method == 'onehot':
            var = list(set(self.x_cols_new) - set(self.additional_persistent_fields))
            x_cat_train = self.model.estimator.get_params()['preprocessor'].get_params()['cat'].fit_transform(self.train_data[self.cat_cols])
            train_cat = pd.DataFrame.sparse.from_spmatrix(x_cat_train,columns=self.cat_cols_trans).reset_index()
            self.df_fet_train = pd.concat([self.train_data[self.cont_cols].reset_index(),train_cat],axis=1)

            shap_df = pd.DataFrame(columns=self.x_cols_new,data = self.shap_values)
            shap_df[self.additional_persistent_fields] = self.train_data.reset_index()[self.additional_persistent_fields]
            self.df_fet_train[self.additional_persistent_fields] = self.train_data.reset_index()[self.additional_persistent_fields]

            fet_values = pd.melt(self.df_fet_train[self.additional_persistent_fields+var], 
                        id_vars=self.additional_persistent_fields, value_vars=var)
            fet_values = fet_values['value']
            fet_values.columns = ['value']

            shap_values_df = pd.melt(shap_df, id_vars=self.additional_persistent_fields, value_vars=var)
            shap_values_df = shap_values_df.rename(columns={'variable':'feature','value':'shap_value'})

            df_fets = pd.concat([shap_values_df,fet_values],axis=1)

        elif self.encode_method == 'label':
            var = list(set(self.x_cols) - set(self.additional_persistent_fields))
            shap_df = pd.DataFrame(columns=self.x_cols,data = self.shap_values)
            shap_df[self.additional_persistent_fields] = self.train_data.reset_index()[self.additional_persistent_fields]

            fet_values = pd.melt(self.train_data[self.additional_persistent_fields+var], 
                                id_vars=self.additional_persistent_fields, value_vars=var)
            fet_values = fet_values['value']
            fet_values.columns = ['value']

            shap_values_df = pd.melt(shap_df, id_vars=self.additional_persistent_fields, value_vars=var)
            shap_values_df = shap_values_df.rename(columns={'variable':'feature','value':'shap_value'})

            df_fets = pd.concat([shap_values_df,fet_values],axis=1)


        outdir = './%s/%s'%(self.model_repo, self.model_desc)
        if not os.path.exists(outdir):
            os.mkdir(outdir)


        self.persistent( df_fets, 'shap_train.csv')
        
        return df_fets


    def generate_shap_test(self):
        """For regression problem, generate shap values for test set and append it with predictions"""
        if self.encode_method == 'onehot':
            var = list(set(self.x_cols_new) - set(self.additional_persistent_fields))
            
            x_cat_test = self.model.estimator.get_params()['preprocessor'].get_params()['cat'].transform(self.test_data[self.cat_cols])
            test_cat = pd.DataFrame.sparse.from_spmatrix(x_cat_test,columns=self.cat_cols_trans)
            self.df_fet_test = pd.concat([self.test_data[[self.y_col]+self.additional_persistent_fields+self.cont_cols].reset_index(),test_cat],axis=1)
            
            shap_df = pd.DataFrame(columns=self.x_cols_new,data = self.test_shap_values)
            shap_df[self.additional_persistent_fields] = self.test_data.reset_index()[self.additional_persistent_fields]

            melt_df = self.df_fet_test
            
            melt_df['pred'] = self.y_pred
            fet_values = pd.melt(melt_df[self.additional_persistent_fields+var+[self.y_col,'pred']], 
                                id_vars=self.additional_persistent_fields+[self.y_col,'pred'], value_vars=var)

            fet_values = fet_values[[self.y_col,'pred','value']]
            fet_values.columns = ['actual','pred','value']
        
            shap_values_df = pd.melt(shap_df, id_vars=self.additional_persistent_fields, value_vars=var)
            shap_values_df = shap_values_df.rename(columns={'variable':'feature','value':'shap_value'})
        
            res = pd.concat([shap_values_df,fet_values],axis=1)

        elif self.encode_method == 'label':
            var = list(set(self.x_cols) - set(self.additional_persistent_fields))
            shap_df = pd.DataFrame(columns=self.x_cols,data = self.test_shap_values)
            shap_df[self.additional_persistent_fields] = self.test_data.reset_index()[self.additional_persistent_fields]

            melt_df = self.test_data.copy()
            melt_df['pred'] = self.y_pred
            fet_values = pd.melt(melt_df[self.additional_persistent_fields+var+[self.y_col,'pred']], 
                                id_vars=self.additional_persistent_fields+[self.y_col,'pred'], value_vars=var)

            fet_values = fet_values[[self.y_col,'pred','value']]
            fet_values.columns = ['actual','pred','value']
        
            shap_values_df = pd.melt(shap_df, id_vars=self.additional_persistent_fields, value_vars=var)
            shap_values_df = shap_values_df.rename(columns={'variable':'feature','value':'shap_value'})
        
            res = pd.concat([shap_values_df,fet_values],axis=1)

        outdir = './%s/%s'%(self.model_repo, self.model_desc)
        if not os.path.exists(outdir):
            os.mkdir(outdir)

        self.persistent(res, 'shap_test.csv')
        
        return res


    def summary_plot(self):
        """summary plot for pure lightgbm"""
        import shap 
        shap.initjs()
        if self.encode_method == 'onehot':
            return shap.summary_plot(self.shap_values, self.df_fet_train[self.x_cols_new])
        elif self.encode_method == 'label':
            return shap.summary_plot(self.shap_values, self.train_data[self.x_cols])


    def tree_viz(self, tree_index):
        """visualize nth tree for pure lightgbm"""
        return lightgbm.create_tree_digraph(self.model.estimator.get_params()['ml_model'], tree_index= tree_index, show_info=['split_gain'])
    
    def shap_plot_by_index(self, plot_index):
        """Plot shap values for nth observation"""
        import shap 
        shap.initjs()
        if type(self.treeExplainer.expected_value) == list:
            self.expected_value = self.treeExplainer.expected_value[0]
        else: 
            self.expected_value = self.treeExplainer.expected_value
        return shap.force_plot(self.expected_value, self.shap_values[plot_index], 
                               self.train_data[self.x_cols].iloc[plot_index])


    def force_plot(self):
        """Plot shap values for all observations"""
        import shap 
        shap.initjs()
        if type(self.treeExplainer.expected_value) == list:
            self.expected_value = self.treeExplainer.expected_value[0]
        else: 
            self.expected_value = self.treeExplainer.expected_value
        if self.encode_method == 'onehot':
            return shap.force_plot(self.expected_value, self.shap_values, self.df_fet_train[self.x_cols_new])
        elif self.encode_method == 'label':
            return shap.force_plot(self.expected_value, self.shap_values, self.train_data[self.x_cols])











class regressionExperiment(ImpExperiment):
    def calc_metrics(self):
        """Regression Metrics"""
        return pd.DataFrame.from_dict({
            'R2':[r2_score(self.y_true, self.y_pred)],
            #'MAPE': [np.mean(np.array([np.abs(self.y_true[i] - self.y_pred[i])/self.y_true[i] for i in range(len(self.y_true)) if self.y_true[i] >0]))],
            'MAE':[mean_absolute_error(self.y_true, self.y_pred)],
            'MSE': [mean_squared_error(self.y_true, self.y_pred)],
            'RMSE': [np.sqrt(mean_squared_error(self.y_true, self.y_pred))]
        })

    
class classificationExperiment(ImpExperiment):
    """Basic Classification Experiments"""
    def configuration(self, **parameter_dict):
        super(classificationExperiment, self).configuration(**parameter_dict)
        self.average_strategy =  parameter_dict['average_strategy']
        
    def test_and_save(self):
            self.y_true, self.y_pred = self.test_y, self.model.predict(self.test_x)

            # result persistent
            self.persistent(self.calc_result(self.test_data, self.y_true.values.flatten(), self.y_pred), 'test/test.csv')
            # metrics persistent
            self.persistent(self.calc_metrics(self.average_strategy), 'test/metrics.csv')
    
    def multiclass_roc_auc_score(self, truth, pred):
        lb = LabelBinarizer()
        lb.fit(truth)
        truth = lb.transform(truth)
        pred = lb.transform(pred)
        return roc_auc_score(truth, pred)
    
    def report(self):    
        report = classification_report(self.y_true, self.y_pred, output_dict=True)
        report = pd.DataFrame.from_dict(report).T
        report['support']['accuracy'] = len(self.y_pred)
        return report

    def calc_metrics(self, average_strategy):
        """Classification Metrics
        """
        try:
            return pd.DataFrame.from_dict({
                'ACCURACY': [accuracy_score(self.y_true, self.y_pred)],
                'AUC_ROC': [self.multiclass_roc_auc_score(self.y_true, self.y_pred)],
                'RECALL':[recall_score(self.y_true, self.y_pred, average=average_strategy)],
                'PRECISION': [precision_score(self.y_true, self.y_pred, average=average_strategy)],
                'F1_SCORE': [f1_score(self.y_true, self.y_pred, average=average_strategy)],
                #'CLASSIFICATION_REPORT': [classification_report(self.y_true, self.y_pred, output_dict=True)]
            })
        except:
            return pd.DataFrame.from_dict({
                'ACCURACY': [accuracy_score(self.y_true, self.y_pred)],
                'RECALL':[recall_score(self.y_true, self.y_pred, average=average_strategy)],
                'PRECISION': [precision_score(self.y_true, self.y_pred, average=average_strategy)],
                'F1_SCORE': [f1_score(self.y_true, self.y_pred, average=average_strategy)],
                #'CLASSIFICATION_REPORT': [classification_report(self.y_true, self.y_pred, output_dict=True)]
            })
        

class NNRegressionExperiment(regressionExperiment):
    """Fully Connected Neural Networks regressor"""
    def configuration(self, **parameter_dict):
        
        super(NNRegressionExperiment, self).configuration(**parameter_dict)
        self.batch_size = parameter_dict['batch_size']
        self.validation = parameter_dict['validation']
        self.patience = parameter_dict['patience']
        self.X_validation, self.y_validation = parameter_dict['validation'][self.x_cols], parameter_dict['validation'][self.y_col].values.flatten()
        self.model_param_list = {'patience':self.patience,'batch_size':self.batch_size}
        
        self.estimator = parameter_dict['estimator']
        self.estimator.set_params(**parameter_dict['estimator_param'])

        self.model = vian_model.FCModel(parameter_dict['model_repo'], parameter_dict['desc'],**self.model_param_list)
        self.model.set_estimator(self.estimator)

    def train(self):
        self.model.train(self.train_x, self.train_y, self.X_validation, self.y_validation)

    def test_and_save(self): 
        self.y_true, self.y_pred = self.test_y.values.flatten(), self.model.predict(self.test_x).flatten()
        self.persistent(self.calc_result(self.test_data, self.y_true, self.y_pred), 'test/test.csv')
        self.persistent(self.calc_metrics(), 'test/metrics.csv')


class NNClassificationExperiment(classificationExperiment):
    """Fully Connected Neural Networks Classifier"""
    def configuration(self, **parameter_dict):
        super(NNClassificationExperiment, self).configuration(**parameter_dict)
        self.average_strategy =  parameter_dict['average_strategy']
        self.batch_size = parameter_dict['batch_size']
        self.validation = parameter_dict['validation']
        self.patience = parameter_dict['patience']
        self.X_validation, self.y_validation = parameter_dict['validation'][self.x_cols], parameter_dict['validation'][self.y_col].values.flatten()
        self.model_param_list = {'patience':self.patience,'batch_size':self.batch_size}

        estimator = parameter_dict['estimator']
        estimator.set_params(**parameter_dict['estimator_param'])

        self.model = vian_model.FCModel(parameter_dict['model_repo'], parameter_dict['desc'],**self.model_param_list)
        self.model.set_estimator(estimator)

    def train(self):
        self.model.train(self.train_x, self.train_y, self.X_validation, self.y_validation)

    def test_and_save(self): 
        self.y_true, self.y_pred = self.test_y.values.flatten(), self.model.predict(self.test_x).flatten()
        self.persistent(self.calc_result(self.test_data, self.y_true, self.y_pred), 'test/test.csv')
        self.persistent(self.calc_metrics(self.average_strategy), 'test/metrics.csv')  
    
    
class NLPClassificationExperiment(classificationExperiment): 
    """NLP Classifier"""
    def configuration(self, **parameter_dict):
        try:
            self.configuration_check(['estimator', 'model_repo', 'desc', 'train', 'test', 'target', 'criterion', 'additional_persistent_fields'], **parameter_dict)
        except AssertionError:
            tb = sys.exc_info()[2]
            raise AssertionError('configuration error').with_traceback(tb)

        estimator = parameter_dict['estimator']
        estimator.set_params(**parameter_dict['estimator_param'])
        
        self.model_repo = parameter_dict['model_repo']
        self.model_desc = parameter_dict['desc']

#         self.validation = parameter_dict['validation']
        self.criterion = parameter_dict['criterion']
        self.model_param_list = {'criterion':self.criterion}
        self.average_strategy =  parameter_dict['average_strategy']
        self.model = vian_model.NLPClassificationModel(parameter_dict['model_repo'], parameter_dict['desc'], **self.model_param_list)
        self.model.set_estimator(estimator)
      
        
        self.x_cols = parameter_dict['text_col']
        self.y_col = parameter_dict['target']
        
        self.train_data = parameter_dict['train']
        
        self.train_x = self.train_data[self.x_cols]
        self.train_y = self.train_data[self.y_col]
        
        # the predict should accept raw data
        self.test_data = parameter_dict['test']
        self.test_x = self.test_data[self.x_cols]
        self.test_y = self.test_data[self.y_col]

        self.additional_persistent_fields = parameter_dict['additional_persistent_fields']
        
        
    def train(self):
        self.model.train(self.train_x, self.train_y)

    def test_and_save(self):
        self.y_true, self.y_pred = self.test_y, self.model.predict(self.test_x)
        self.persistent(self.calc_result(self.test_data, self.y_true, self.y_pred), 'test/test.csv')
        self.persistent(self.calc_metrics(average_strategy = self.average_strategy), 'test/metrics.csv')


class MyLEncoder():
    def __init__(self,columns = None):
        self.columns = columns # array of column names to encode
        self.encoders = [LabelEncoder() for _ in range(len(self.columns))]

    def fit(self,X,y=None):
        return self # not relevant here

    def trans(self,X):
        '''
        Transforms columns of X specified in self.columns using
        LabelEncoder(). If no columns specified, transforms all
        columns in X.
        '''
        X = pd.DataFrame(X)
        X.columns = self.columns
        output = X.copy()
        #if self.columns is not None:
        for i in range(len(self.columns)):
            output[self.columns[i]] = self.encoders[i].fit_transform(output[self.columns[i]])
        return output.values

    def fit_transform(self,X,y=None):
        return self.fit(X,y).trans(X)

    def transform(self,X,y=None):
        X = pd.DataFrame(X)
        X.columns = self.columns
        output = X.copy()
        #if self.columns is not None:
        for i in range(len(self.columns)):
            output[self.columns[i]] = self.encoders[i].transform(output[self.columns[i]])
        return output.values

def build_estimator(cat_cols, cont_cols, ml_model, encode_method = 'onehot'):
    """ Build model pipeline
    """
    transformer = []
    numeric_transformer = Pipeline(steps=[
                            ('imputer', SimpleImputer(strategy='median')),
                            ('scaler', StandardScaler())
    ])
    transformer.append(('num', numeric_transformer, cont_cols))
    
    if encode_method == 'onehot':
        categorical_transformer = Pipeline(steps=[
                            ('imputer', SimpleImputer(strategy='most_frequent')),
                            ('oneHot', OneHotEncoder(handle_unknown='ignore')),
                                ])
        transformer.append(('cat', categorical_transformer, cat_cols))
    elif encode_method == 'label':
        categorical_transformer = Pipeline(steps=[
                            ('imputer', SimpleImputer(strategy='most_frequent')),
                            ('label',MyLEncoder(columns=cat_cols))
                                ])
        transformer.append(('cat', categorical_transformer, cat_cols))

    preprocessor = ColumnTransformer(transformer)
    return Pipeline(steps=[('preprocessor', preprocessor),
                            ('ml_model', ml_model)])
    
def create_fc_model(n_features = 100):
    model = Sequential()
    model.add(Dense(512, input_dim=n_features , kernel_initializer='RandomNormal',activation = 'relu'))
    model.add(BatchNormalization(axis=-1, momentum=0.99, epsilon=0.001, center=True, scale=True, beta_initializer='zeros', gamma_initializer='ones', moving_mean_initializer='zeros', moving_variance_initializer='ones', beta_regularizer=None, gamma_regularizer=None, beta_constraint=None, gamma_constraint=None))
    model.add(Dropout(0.1))
    model.add(Dense(32,activation='relu'))
    model.add(BatchNormalization(axis=-1, momentum=0.99, epsilon=0.001, center=True, scale=True, beta_initializer='zeros', gamma_initializer='ones', moving_mean_initializer='zeros', moving_variance_initializer='ones', beta_regularizer=None, gamma_regularizer=None, beta_constraint=None, gamma_constraint=None))
    model.add(Dense(256,activation= 'sigmoid'))
    model.add(Dropout(0.1))
    model.add(Dense(64,activation='relu'))
    model.add(BatchNormalization(axis=-1, momentum=0.99, epsilon=0.001, center=True, scale=True, beta_initializer='zeros', gamma_initializer='ones', moving_mean_initializer='zeros', moving_variance_initializer='ones', beta_regularizer=None, gamma_regularizer=None, beta_constraint=None, gamma_constraint=None))
    model.add(Dense(16,activation='relu'))
    model.add(Dense(1,activation='relu'))
    model.compile(optimizer=tf.keras.optimizers.Adam(lr=0.001), loss= 'mae')
    return model


def publish_test_result(df_test_result, df_test_result_metrics, test_desc, term_id, time_stamp, project='client-257322', schema='atm_forecasting'):
    """ publish test result and metrics data to bigquery
    """
    user = os.environ['USER']
    assert 'term_id' in df_test_result.columns, 'expect column [term_id] from the test result data'

    df_test_result['test_description'] = test_desc
    df_test_result['timestamp'] = time_stamp
    df_test_result['user'] = user
    df_test_result['term_id'] = term_id

    df_test_result_metrics['test_description'] = test_desc
    df_test_result_metrics['timestamp'] = time_stamp
    df_test_result_metrics['user'] = user

    # TODO: temporary convert it to int64 for now, we'll need a strict type checking here
    df_test_result['actual'] = df_test_result['actual'].astype('int64')
    print(df_test_result)
    df_test_result.to_gbq('%s.test_result'%schema, project, if_exists='append') 
    df_test_result_metrics.to_gbq('%s.test_result_metrics'%schema, project, if_exists='append')

def publish_test_result_detail_only(df_test_result, test_desc, time_stamp, project='client-257322', schema='atm_forecasting'):
    """ publish test result data to bigquery
    """
    user = os.environ['USER']
    assert 'term_id' in df_test_result.columns, 'expect column [term_id] from the test result data'

    df_test_result['test_description'] = test_desc
    df_test_result['timestamp'] = time_stamp
    df_test_result['user'] = user
    df_test_result['actual'] = df_test_result['actual'].astype('int64')
    df_test_result['REAL_DATE'] = df_test_result['REAL_DATE'].astype(str)
    df_test_result.to_gbq('%s.test_result'%schema, project, if_exists='append') 

def publish_test_result_w_fet(df_test_result, 
                              fet_cols,
                              test_desc, 
                              time_stamp, 
                              project='client-257322', 
                              schema='atm_forecasting'):
    """ publish test result data with features to bigquery
    """
    user = os.environ['USER']
    assert 'term_id' in df_test_result.columns, 'expect column [term_id] from the test result data'
    for fc in fet_cols:
        assert fc in df_test_result.columns, 'expect column [%s] from the test result data'%fc

    df_test_result['test_description'] = test_desc
    df_test_result['timestamp'] = time_stamp
    df_test_result['user'] = user
    df_test_result['actual'] = df_test_result['actual'].astype('int64')
    df_test_result['REAL_DATE'] = df_test_result['REAL_DATE'].astype(str)

    df_test_result.to_gbq('%s.test_result_w_fet'%schema, project, if_exists='append') 

def data_split(df,method='random',split_time=None,dt_col= None,test_size=0.2,random_state=42):
    """ train test split
    """
    if method == 'random':
        df_train, df_test = train_test_split(df, test_size=test_size, random_state=random_state)
        return df_train, df_test
    elif method == 'time':
        df[dt_col] = pd.to_datetime( df[dt_col] )
        df_train = df.loc[df[dt_col] < pd.to_datetime(split_time)]
        df_test = df.loc[(df[dt_col] >= pd.to_datetime(split_time))]
        return df_train, df_test
    else: 
        raise ValueError('split method must be "random" or "time"')

def get_n_features(df,cont_cols,cat_cols):
    """ get input n features size for FC model
    """
    n_features = len(cont_cols)
    for c in cat_cols:
        n_features+= len(df[c].unique())
        if df[c].isna().any():
            n_features -= 1
    return n_features

def myloss(y_true, y_pred):
    y_true = tf.cast(y_true,tf.float32)
    y_pred = tf.cast(y_pred,tf.float32)
    idx  = tf.not_equal(y_true,0)
    y_true = tf.boolean_mask(y_true, idx)
    y_pred = tf.boolean_mask(y_pred, idx)
    idx  = tf.not_equal(y_pred,0)
    y_true = tf.boolean_mask(y_true, idx)
    y_pred = tf.boolean_mask(y_pred, idx)
    loss = tf.cond(tf.equal(tf.shape(y_pred)[0], 0), lambda: tf.constant(0, dtype=tf.float32), lambda: tf.keras.backend.sum(tf.abs(y_true-y_pred))/tf.keras.backend.sum((y_true+y_pred)))
    return loss 

def load_data(dataset,from_local=False):
    """ Load sample data
    """
    if from_local:
        DATA_PATH = dataset
    else:
        this_dir, this_filename = os.path.split(__file__)
        DATA_PATH = os.path.join(this_dir, "sample_data", dataset+'.csv')
    try:
        df = pd.read_csv(DATA_PATH)
        return df
    except FileNotFoundError:
        print("Dataset Not Found")
